#define _POSIX_C_SOURCE 200112L

#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>

#define RTCPROTO_IMPLEMENTATION
#include "rtcproto.h"

#include "sim.h"

typedef struct Synchronizer
{
    struct timespec last;
    struct timespec delta;
    struct timespec period;
} Synchronizer;

void synchronizer_init(Synchronizer* sync, time_t period_nsec)
{
    /* Set last synchronized time as the current time */
    clock_gettime(CLOCK_REALTIME, &sync->last);

    /* Store period */
    sync->period.tv_sec = 0;
    sync->period.tv_nsec = period_nsec;
}

void synchronizer_wait(Synchronizer* sync)
{
    struct timespec now;
    struct timespec next;
    struct timespec want;

    /* Get current time */
	clock_gettime(CLOCK_REALTIME, &now);

    /* Get synchronization timepoint */
    next = rtcp_timespec_add(&sync->last, &sync->period);

    /* If the timepoint was not yet passed, wait until then */
    if(rtcp_timespec_less(&now, &next)){
        want = rtcp_timespec_diff(&now, &next);
        nanosleep(&want, NULL);
    }
}

void synchronizer_step(Synchronizer* sync)
{
    struct timespec after;

    /* Note delta time */
	clock_gettime(CLOCK_REALTIME, &after);
    sync->delta = rtcp_timespec_diff(&sync->last, &after);

    /* Set the timepoint after the wait as the last */
    sync->last = after;
}

typedef struct TrophsBuffer
{
	Trophs trophs[2];
	Trophs* active;
	Trophs* passive;
	int i;
} TrophsBuffer;

typedef struct Pointers
{
	void** active;
	void** passive;
} Pointers;

PT_THREAD(trophs_bigstep(rtcp_pt_t* pt, void* state, struct timespec* end))
{
	TrophsBuffer* tbuf = *(TrophsBuffer**)state;
	int i;

	PT_BEGIN(pt);

	PT_LOOP(tbuf->i = 0, tbuf->i < 20000, i, 10){

		/* Step the cellular automata */
		trophs_step(tbuf->passive);
	
		PT_LOOP_INCR(pt, end, ++tbuf->i, i)
	}

	PT_END(pt);
}

PT_THREAD(trophs_copy(rtcp_pt_t* pt, void* state, struct timespec* end))
{
	TrophsBuffer* tbuf = *(TrophsBuffer**)state;
	
	PT_BEGIN(pt);

	(void)end;

	/* Copying only the cell state is enough here */
	memcpy(tbuf->passive->cells, tbuf->active->cells,
		tbuf->active->off_new + tbuf->active->bytes);

	PT_END(pt);
}

PT_THREAD(trophs_swap(rtcp_pt_t* pt, void* state, struct timespec* end))
{
	(void)end;
	Pointers* tbuf = (Pointers*)state;
	void* swap = NULL;

	PT_BEGIN(pt);
	
	memcpy(&swap, tbuf->active, sizeof(void*));
	memcpy(tbuf->active, tbuf->passive, sizeof(void*));
	memcpy(tbuf->passive, &swap, sizeof(void*));

	PT_END(pt);
}

const int TROPH_WIDTH = 32;

int main(int argc, char** argv)
{
	Synchronizer sync;
	TrophsBuffer tbuf;
	rtcp_sched_t sched;
	rtcp_stats_t stats;
	int ret;
    (void)argc;
    (void)argv;

	/* Set stdin to non-blocking mode */
	fcntl(0, F_SETFL, O_NONBLOCK);

	/* Set up double buffer */
	tbuf.active = &tbuf.trophs[0];
	tbuf.passive = &tbuf.trophs[1];
	trophs_malloc(tbuf.active, TROPH_WIDTH);
	trophs_malloc(tbuf.passive, TROPH_WIDTH);
	trophs_initialize(tbuf.active, (float[4]){ 0.27f, 0.01f, 0.12f, 0.06f });
	trophs_initialize(tbuf.passive, (float[4]){ 0.27f, 0.01f, 0.12f, 0.06f });
	trophs_random(tbuf.active);
	trophs_random(tbuf.passive);

	/* Set up scheduler */
	ret = rtcp_sched_malloc(&sched);
	rtcp_assert_always(ret == 0, "failed to allocate");

	/* Register function ahead of time for convenience in retrieving statistics */
	ret = rtcp_sched_func_get_assoc(&sched,
		trophs_bigstep, sizeof(TrophsBuffer*), NULL);
	rtcp_assert_always(ret == 0, "failed to allocate");
	ret = rtcp_sched_func_get_assoc(&sched,
		trophs_swap, sizeof(Pointers), NULL);
	rtcp_assert_always(ret == 0, "failed to allocate");
	ret = rtcp_sched_func_get_assoc(&sched,
		trophs_copy, sizeof(TrophsBuffer*), NULL);
	rtcp_assert_always(ret == 0, "failed to allocate");

	/* Set syncrhonization interval */
	synchronizer_init(&sync, 100 * 1000000);

	for(;;){
		
		/* Wait using the synchronizer, then step its state */
		synchronizer_wait(&sync);
		synchronizer_step(&sync);
		
		/* Step the cellular automata */
		trophs_step(tbuf.active);

		/* Print */
		trophs_show(tbuf.active);
		rtcp_sched_func_stats(&sched, trophs_bigstep, &stats);
		printf("entries: %d, total: %lu.%lu, max_lc: %u, max_over: %.3f ms\n",
			stats.entries, stats.total.tv_sec, stats.total.tv_nsec,
			stats.max_lc, stats.max_over.tv_nsec / (float)1e6);
		printf("size/capa/targetcapa: sched.calls:  %d/%d/%d\n",
			sched.calls.size, sched.calls.capacity, sched.calls.targetcapa);
		printf("size/capa/targetcapa: sched.assoc:  %d/%d/%d\n",
			sched.assoc.size, sched.assoc.capacity, sched.assoc.targetcapa);
			
		rtcp_assoc_t* assoc = rtcp_hmap_get(&sched.assoc, trophs_bigstep);
		printf("size/capa/targetcapa: assoc.locals: %d/%d/%d (trophs_bigstep)\n",
			assoc->locals.size, assoc->locals.capacity, assoc->locals.targetcapa);
		assoc = rtcp_hmap_get(&sched.assoc, trophs_swap);
		printf("size/capa/targetcapa: assoc.locals: %d/%d/%d (trophs_swap)\n",
			assoc->locals.size, assoc->locals.capacity, assoc->locals.targetcapa);
		assoc = rtcp_hmap_get(&sched.assoc, trophs_copy);
		printf("size/capa/targetcapa: assoc.locals: %d/%d/%d (trophs_copy)\n",
			assoc->locals.size, assoc->locals.capacity, assoc->locals.targetcapa);

		printf("synchronizer delta: %.3f ms\n", sync.delta.tv_nsec / (float)1e6);

		/* Attempt to retrieve a character from stdin, or EOF */
		char c = getchar();

		/* If a newline was received */
		if(c != EOF && c == '\n'){
			
			/* Queue up a big step, copyback, and a buffer swap */
			TrophsBuffer* tbufptr = &tbuf;
			ret = rtcp_sched_mqueue(&sched, sync.last, trophs_bigstep, tbufptr);
			ret = rtcp_sched_mqueue(&sched, sync.last, trophs_swap,
				((Pointers){ (void**)&tbuf.active, (void**)&tbuf.passive }));
			ret = rtcp_sched_mqueue(&sched, sync.last, trophs_copy, tbufptr);
		}

		/* Call the scheduler */
		rtcp_sched_until(&sched, rtcp_timespec_add(&sync.last, &sync.period));

		/* Tick & shrink the scheduler */
		rtcp_sched_tick(&sched);
		rtcp_sched_shrink(&sched);
	}

    return 0;
}
