#include "sim.h"

void trophs_malloc(Trophs* trophs, int width)
{
    rtcp_assert_always(width >= 0, "trophs_initialize: width negative\n");

    //Check that the grid size is even so that it may be compressed 4x
    rtcp_assert_always(width % 2 == 0, "trophs_initialize: width not even\n");

    //Store parameters
    trophs->width = width;

    //Calculate byte offsets for areas starting from offsetof(Trophs, cells)
    trophs->bytes = trophs->width * trophs->width;
    trophs->off_comp = 0;
    trophs->off_old = trophs->off_comp;
    trophs->off_new = trophs->off_old + trophs->bytes;

	//Allocate memory area
	trophs->cells = malloc(trophs->off_new + trophs->bytes);
}

void trophs_free(Trophs* trophs)
{
	//Free memory are
	if(trophs->cells){ free(trophs->cells); }
}

void trophs_initialize(Trophs* trophs, float params[TROPH_PARAM_COUNT])
{
    //Initialize values
    trophs->grid = 0;

    //Neighborhood offsets
    //[0, 4): Von-Neumann, [0, 8): Moore
    memcpy(trophs->neigh,
        &(int[8][2]){ { -1,  0 }, { +1,  0 }, { 0,  -1 }, { 0,  +1 },
                      { -1, -1 }, { +1, -1 }, { -1, +1 }, { +1, +1 }, },
        sizeof(int[8][2]));

    //Offset neighborhoods by the width so that
    //remainder operators ('%') will acts as modulus
    for(int i = 0; i < 4; ++i){
        trophs->neigh[i][0] += trophs->width;
        trophs->neigh[i][1] += trophs->width;
    }

    //Seed romu generator
    trophs->romu_trio32[0] = 0x4761fb1a;
    trophs->romu_trio32[1] = 0x94b3cd22;
    trophs->romu_trio32[2] = 0xb3dd53a6;

    //Autotroph & heterotroph, reproduction & death rate parameters
    for(int i = 0; i < TROPH_PARAM_COUNT; ++i){
        trophs->params[i] = params[i];
    }
    trophs->hete_hunt = 2;

    //Initialize cells to dead
    memset(trophs->cells + trophs->off_old, TROPH_DEAD, trophs->bytes);
}

void trophs_random(Trophs* trophs)
{
    //Pointer to cell
    uint8_t* old = trophs->cells +
        (trophs->grid ? trophs->off_new : trophs->off_old);

	//Initialize cells randomly
	for(int i = 0; i < trophs->bytes; ++i){
		
    	uint32_t rand = romu_trio32_random(trophs->romu_trio32) % 100;
		
		old[i] = TROPH_DEAD;
		if(rand <= 70){ old[i] = TROPH_AUTO; }
		if(rand > 70 && rand < 100){ old[i] = TROPH_HETE; }
	}
}

void trophs_show(Trophs* trophs)
{
    //Pointer to cell
    uint8_t* old = trophs->cells +
        (trophs->grid ? trophs->off_new : trophs->off_old);

    for(int i = 0; i < trophs->width; ++i){
        for(int j = 0; j < trophs->width; ++j){
            switch(old[i * trophs->width + j]){
            case TROPH_DEAD: { printf(ESC_RST "  " ESC_RST); } break;
            case TROPH_AUTO: { printf(ESC_GRN "# " ESC_RST); } break;
            case TROPH_HETE: { printf(ESC_RED "x " ESC_RST); } break;
            default: { exit(1); } break;
            }
        }
        printf("\n");
    }
}

void trophs_step(Trophs* trophs)
{
    //Pointer to cell
    uint8_t* new = trophs->cells +
        (trophs->grid ? trophs->off_old : trophs->off_new);
    uint8_t* old = trophs->cells +
        (trophs->grid ? trophs->off_new : trophs->off_old);

    //Stochastic parameters
    float params[TROPH_PARAM_COUNT];
    for(int i = 0; i < TROPH_PARAM_COUNT; ++i){
        params[i] = trophs->params[i];
    }

    int width = trophs->width;

    //Fill new grid with dead cells initially
    memset(new, TROPH_DEAD, trophs->bytes);

    //Autotroph to heterotroph, heterotroph to dead
    for(int i = 0; i < width; ++i){ int row_off = i * width;
    for(int j = 0; j < width; ++j){

        switch(old[row_off + j]){
        case TROPH_AUTO: {

            uint16_t moore = 0;
            for(int k = 0; k < 8; ++k){
                moore += 0x1 << (
                    (old[
                        ((i + trophs->neigh[k][0]) % width) * width +
                        ((j + trophs->neigh[k][1]) % width)
                    ]) * 4
                );
            }
            
            //Autotroph to heterotroph
            if(
                ((moore & 0x0f00) >> 8) >= (uint8_t)trophs->hete_hunt &&
                romu_trio32_random(trophs->romu_trio32) <
                    (uint32_t)(UINT32_MAX * params[TROPH_HETE_REPRO])
            ){
                old[row_off + j] = 0;
                new[row_off + j] = TROPH_HETE;
            } else {
                new[row_off + j] = TROPH_AUTO;
            }

        } break;
        case TROPH_HETE: {

            //Heterotroph to dead
            new[row_off + j] =
                romu_trio32_random(trophs->romu_trio32) <
                (uint32_t)(UINT32_MAX * params[TROPH_HETE_DEATH]) ?
                TROPH_DEAD : TROPH_HETE;

        } break;
        default: {} break;
        }
    }}

    //Dead to autotroph, autotroph to dead
    for(int i = 0; i < width; ++i){ int row_off = i * width;
    for(int j = 0; j < width; ++j){

        switch(old[row_off + j]){
        case TROPH_AUTO: {

            //Autotroph reproduction
            if(romu_trio32_random(trophs->romu_trio32) <
                (uint32_t)(UINT32_MAX * params[TROPH_AUTO_REPRO])
            ){
                //Get index of random empty neighbor
                int k_off = romu_trio32_random(trophs->romu_trio32) % 4;

                int k = k_off; int kmod = k % 4;
                while(
                    k < k_off + 4 &&
                    old[
                        ((i + trophs->neigh[kmod][0]) % width) * width +
                        ((j + trophs->neigh[kmod][1]) % width)
                    ] != TROPH_DEAD
                ){ ++k; kmod = k % 4; }

                //If there is an empty neighbor, reproduce there
                if(k < k_off + 4){
                    new[
                        ((i + trophs->neigh[kmod][0]) % width) * width +
                        ((j + trophs->neigh[kmod][1]) % width)
                    ] = TROPH_AUTO;
                }
            }

            //Autotroph to dead
            new[row_off + j] =
                romu_trio32_random(trophs->romu_trio32) <
                (uint32_t)(UINT32_MAX * params[TROPH_AUTO_DEATH]) ?
                TROPH_DEAD : TROPH_AUTO;

        } break;
        default: {} break;
        }
    }}

    //Set pointer to currently active cell grid
    trophs->grid = 1 - trophs->grid;
}
